import unittest

from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group
from spetlr.spark import Spark


class TransformerTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.transformer = TestTransformer()
        cls.df = create_dataframe()

    def test_process(self):
        # Process is called with single input
        # Result contain single df
        result = self.transformer.etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])

    def test_process_many(self):
        # Process_many is called with multiple inputs
        # Result contain union of input dfs
        result = self.transformer.etl({"df1": self.df, "df2": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"TestTransformer"}, set(result.keys()))
        self.assertEqual(
            self.df.union(self.df).collect(), result["TestTransformer"].collect()
        )

    def test_dataset_key(self):
        # Set dataset_key to control name of the output
        # Result contain the key 'my_key'
        result = TestTransformer(dataset_key="my_key").etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"my_key"}, set(result.keys()))
        self.assertIs(self.df, result["my_key"])

    def test_dataset_output_key(self):
        # Set dataset_output_key to control name of the output
        # Result contain the key 'my_key'
        result = TestTransformer(dataset_output_key="my_key").etl({"df": self.df})
        self.assertEqual(1, len(result))
        self.assertEqual({"my_key"}, set(result.keys()))
        self.assertIs(self.df, result["my_key"])

    def test_single_dataset_input_key(self):
        # Set dataset_input_keys to control what dfs in dataset to handle
        # Process is called with single input
        # Result contain two keys due to 'df2' not being consumed
        # when not in dataset_input_keys
        result = TestTransformer(dataset_input_keys=["df1"]).etl(
            {"df1": self.df, "df2": self.df}
        )
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df2"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])
        self.assertIs(self.df, result["df2"])

    def test_multiple_dataset_input_keys(self):
        # Set dataset_input_keys to control what dfs in dataset to handle
        # Process_many is called with multiple inputs
        # Result contain two keys due to 'df3' not being consumed
        # when not in dataset_input_keys
        result = TestTransformer(dataset_input_keys=["df1", "df2"]).etl(
            {"df1": self.df, "df2": self.df, "df3": self.df}
        )
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df3"}, set(result.keys()))
        self.assertEqual(
            self.df.union(self.df).collect(), result["TestTransformer"].collect()
        )
        self.assertIs(self.df, result["df3"])

    def test_not_consume_inputs(self):
        # Set consume_inputs to False to control dataset keys will no be consumed
        # in the output
        # Result contain two keys due to 'df' not being consumed
        result = TestTransformer(consume_inputs=False).etl({"df": self.df})
        self.assertEqual(2, len(result))
        self.assertEqual({"TestTransformer", "df"}, set(result.keys()))
        self.assertIs(self.df, result["TestTransformer"])
        self.assertIs(self.df, result["df"])


class TestTransformer(Transformer):
    def process(self, df: DataFrame) -> DataFrame:
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        assert len(datasets) == 2
        df1, df2 = list(datasets.values())
        return df1.union(df2)


def create_dataframe():
    data = [(1, "1"), (2, "2"), (3, "3")]

    schema = StructType(
        [
            StructField("id", IntegerType(), False),
            StructField("text", StringType(), False),
        ]
    )

    return Spark.get().createDataFrame(data=data, schema=schema)


if __name__ == "__main__":
    unittest.main()



from typing import List

from spetlr.delta import DeltaHandle
from spetlr.etl import EtlBase, Orchestrator
from spetlr.etl.extractors import IncrementalExtractor, SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.UpsertLoader import UpsertLoader
from spetlr.exceptions import MissingUpsertJoinColumns
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)


class EhToDeltaSilverOrchestrator(Orchestrator):
    """
    This class has been designed to carry out the ETL task
    of unpacking and transforming bronze eventhub data to the silver layer.

    Parameters:

    dh_source: DeltaHandle for the source delta table (bronze)
    dh_target: DeltaHandle for the target delta table (silver)
    upsert_join_cols: The columns used for upserting.
    mode: The mode of data load (upsert, append, overwrite).

    Returns:
    Processed datasets of the super Orchestrator class
    """

    def __init__(
        self,
        dh_source: DeltaHandle,
        dh_target: DeltaHandle,
        upsert_join_cols: List[str] = None,
        mode: str = "upsert",
    ):
        super().__init__()
        self.dh_source = dh_source
        self.dh_target = dh_target
        self.mode = mode
        self.upsert_join_cols = upsert_join_cols

        # step 1
        # Extracts the data from the bronze layer
        if mode == "upsert" or mode == "append":
            if mode == "upsert":
                if upsert_join_cols is None:
                    raise MissingUpsertJoinColumns
            self.extract_from(
                IncrementalExtractor(
                    handle_source=self.dh_source,
                    handle_target=self.dh_target,
                    time_col_source="EnqueuedTimestamp",
                    time_col_target="EnqueuedTimestamp",
                )
            )

        else:
            self.extract_from(SimpleExtractor(dh_source, "EhDeltaBronze"))

        # step 2,
        #  - use the target schema to select what to copy from capture files
        #  - anything that is not in the source df is used to unpack the body json
        self.transform_with(EhJsonToDeltaTransformer(target_dh=dh_target))

        # the method filter_with can be used to insert any number of transformers here

        # final step,
        # by default upsert the rows to the silver delta table.
        if mode == "upsert":
            self._loader = UpsertLoader(dh_target, self.upsert_join_cols)
        else:
            self._loader = SimpleLoader(dh_target, mode=self.mode)

        self.load_into(self._loader)

    @classmethod
    def from_tc(cls, dh_source_id: str, dh_target_id: str):
        return cls(
            dh_source=DeltaHandle.from_tc(dh_source_id),
            dh_target=DeltaHandle.from_tc(dh_target_id),
        )

    def filter_with(self, etl: EtlBase):
        """Additional filters to execute before loading."""
        loader = self.steps.pop()

        # the following will fail if additional steps have been added.
        assert self._loader is loader, "unexpected change in etl steps"

        self.transform_with(etl)
        self.load_into(loader)



from abc import abstractmethod
from typing import List, Union

from deprecated import deprecated
from pyspark.sql import DataFrame

from spetlr.etl.types import dataset_group

from .transformer import Transformer


@deprecated(reason="Use Transformer with consume_inputs=False instead")
class TransformerNC(Transformer):
    """If you only want to transform a single input dataframe,
    implement `process`
    If you want to transform a set of dataframes,
    implement `process_many`

    In regards to the etl step, the TransformerNC does NOT CONSUME the inputs
    and ADDs the result of its transformation stage to the dataset dict.
    """

    def __init__(
        self,
        *,
        dataset_input_keys: Union[str, List[str]] = None,
        dataset_output_key: str = None,
    ):
        if isinstance(dataset_input_keys, str):
            dataset_input_keys = [dataset_input_keys]
        else:
            dataset_input_keys = dataset_input_keys

        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=False,
        )

    @abstractmethod
    def process(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    @abstractmethod
    def process_many(self, datasets: dataset_group) -> DataFrame:
        raise NotImplementedError()



from unittest.mock import Mock, patch

from pyspark.sql import DataFrame
from pyspark.sql import functions as f
from spetlrtools.testing import DataframeTestCase

from spetlr.etl import Transformer
from spetlr.etl.extractors import IncrementalExtractor, SimpleExtractor
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.UpsertLoader import UpsertLoader
from spetlr.exceptions import MissingUpsertJoinColumns
from spetlr.orchestrators import EhToDeltaSilverOrchestrator
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)
from spetlr.spark import Spark


class EhToDeltaSilverOrchestratorTests(DataframeTestCase):
    def test_01_default_upsert(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(UpsertLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        upsert_join_cols=["id"],
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_02_input_upsert(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(UpsertLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="upsert",
                        upsert_join_cols=["id"],
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_03_upsert_missing_upsert_cols(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(UpsertLoader, "save", return_value=test_df) as p2:
                    with self.assertRaises(MissingUpsertJoinColumns) as cm:
                        orchestrator = EhToDeltaSilverOrchestrator(
                            dh_source=dh_source_mock,
                            dh_target=dh_target_mock,
                            mode="upsert",
                        )
                        orchestrator.execute()

                        p0.assert_called_once()
                        p1.assert_called_once()
                        p2.assert_called_once()
                        cm.exception = "You must specify upsert_join_cols"

    def test_04_w_transformer(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        class TestFilter(Transformer):
            def process(self, df: DataFrame) -> DataFrame:
                return df.withColumn("Test", f.lit("Test"))

        class TestOrchestrator(EhToDeltaSilverOrchestrator):
            def __init__(self):
                super().__init__(
                    dh_source=dh_source_mock,
                    dh_target=dh_target_mock,
                    upsert_join_cols=["id"],
                )
                self.filter_with(TestFilter())

        with patch.object(
            EhJsonToDeltaTransformer, "process", return_value=test_df
        ) as p1:
            with patch.object(TestFilter, "process", return_value=test_df) as p3:
                orchestrator = TestOrchestrator()
                orchestrator.execute()

                dh_source_mock.read.assert_called_once()
                p1.assert_called_once()
                dh_target_mock.upsert.assert_called_once()
                p3.assert_called_once()

    def test_05_append(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(IncrementalExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="append",
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()

    def test_06_overwrite(self):
        test_df = Spark.get().createDataFrame(
            [], schema="id string, EnqueuedTimestamp timestamp"
        )
        dh_source_mock = Mock()
        dh_source_mock.read = Mock(return_value=test_df)
        dh_target_mock = Mock()
        dh_target_mock.read = Mock(return_value=test_df)

        with patch.object(SimpleExtractor, "read", return_value=test_df) as p0:
            with patch.object(
                EhJsonToDeltaTransformer, "process", return_value=test_df
            ) as p1:
                with patch.object(SimpleLoader, "save", return_value=test_df) as p2:
                    orchestrator = EhToDeltaSilverOrchestrator(
                        dh_source=dh_source_mock,
                        dh_target=dh_target_mock,
                        mode="overwrite",
                    )
                    orchestrator.execute()

                    p0.assert_called_once()
                    p1.assert_called_once()
                    p2.assert_called_once()
